#!/usr/bin/env python3
'''
Dummy experiment.
'''

from hydronaut.experiment import Experiment


class DummyExperiment(Experiment):
    '''
    Dummy experiment.
    '''

    # This is the only method that subclasses are required to implement.
    def __call__(self):
        # Get the Hydra config object from the parent class.
        config = self.config

        # Get the experiment parameters from the config object.
        params = config.experiment.params

        # Return some numerical value based on the parameters. In a real
        # experiment, this would be a validation or test score after training
        # and fitting a model.
        return params.x * params.y * params.z



#!/usr/bin/env python3
'''
MLflow utility functions.
'''

import logging
import os
from contextlib import ExitStack
import pathlib
from types import TracebackType
from typing import Any, Optional, Type

import mlflow
from hydra.core.hydra_config import HydraConfig
from omegaconf import DictConfig, ListConfig

from hydronaut.hydra.omegaconf import get_container
from hydronaut.types import Path


LOGGER = logging.getLogger(__name__)


def _recursively_log_omegaconf_params_with_mlflow(name: str, obj: Any) -> None:
    '''
    Recursively log the Hydra parameters.

    Args:
        name:
            The parameter name.

        obj:
            The parameter object.
    '''
    if isinstance(obj, DictConfig):
        for key, value in obj.items():
            _recursively_log_omegaconf_params_with_mlflow(f'{name}.{key}', value)
    elif isinstance(obj, ListConfig):
        for i, value in enumerate(obj):
            _recursively_log_omegaconf_params_with_mlflow(f'{name}.{i:d}', value)
    else:
        LOGGER.debug('logging parameter: %s = %s', name, obj)
        mlflow.log_param(name, obj)


class MLflowRunner():
    '''
    Context manager to set up an MLflow run from an Omegaconf configuration
    object. When used as a context manager, it will configure the tracking URI,
    experiment, artifact location and run parameters from the configurat object.
    The context will return the MLflow active run.
    '''
    def __init__(self, config: DictConfig, base_dir: Path = None) -> None:
        '''
        Args:
            config:
                The Omegaconf configuration object.
        '''
        self.config = config
        if base_dir is not None:
            base_dir = pathlib.Path(base_dir).resolve()
        self.base_dir = base_dir
        self.exit_stack = ExitStack()
        self.active_run_id = None

    def log_parameters(self) -> None:
        '''
        Log all parameters with MLflow.
        '''
        for name, value in self.config.items():
            _recursively_log_omegaconf_params_with_mlflow(name, value)

    def _set_experiment(self) -> mlflow.entities.Experiment:
        '''
        Set the current experiment from the name in the configuration object.

        Returns:
            An instance of mlflow.entities.Experiment.
        '''
        config = self.config

        exp_name = config.experiment.name
        LOGGER.info('MLflow experiment name: %s', exp_name)

        exp = mlflow.get_experiment_by_name(exp_name)
        if exp is None:
            exp_id = mlflow.create_experiment(name=exp_name)
            exp = mlflow.get_experiment(exp_id)
        mlflow.set_experiment(experiment_id=exp.experiment_id)

        return exp

    def _set_uris(self) -> None:
        '''
        Log and set the MLflow URIs.
        '''
        env_var = 'MLFLOW_TRACKING_URI'
        if self.base_dir and env_var not in os.environ:
            os.environ[env_var] = (self.base_dir / 'mlruns').as_uri()
        tracking_uri = mlflow.get_tracking_uri()
        LOGGER.info('MLflow tracking URI: %s', tracking_uri)

        registry_uri = os.getenv('MLFLOW_REGISTRY_URI', default=tracking_uri)
        if registry_uri:
            mlflow.set_registry_uri(registry_uri)

        LOGGER.info(
            'MLflow registry URI: %s',
            mlflow.get_registry_uri()
        )

    def _get_mlflow_run_data(self, prop, name):
        '''
        Internal function for returning run data values from the active MLflow run.

        Args:
            prop:
                The RunData property to query (metrics, params or tags).

            name:
                The name of the value to return.
        '''
        if prop not in ('metrics', 'params', 'tags'):
            raise ValueError(f'Unrecognized RunData property requested: {prop}')

        if self.active_run_id is None:
            raise RuntimeError(f'Cannot retrieve MLflow {prop} outside of active run.')

        data = mlflow.get_run(self.active_run_id).data

        value = getattr(data, prop).get(name)
        if value is None:
            LOGGER.warning(
                'The active MLflow run (%s) currently has no %s named %s',
                self.active_run_id,
                prop[:-1],  # -1 to remove the s
                name
            )
        return value

    def get_metric(self, name):
        '''
        Return the value of a metric logged during the current run.
        '''
        return self._get_mlflow_run_data('metrics', name)

    def get_param(self, name):
        '''
        Return the value of a parameter logged during the current run.
        '''
        return self._get_mlflow_run_data('params', name)

    def __enter__(self) -> mlflow.ActiveRun:
        config = self.config
        hydra_config = HydraConfig.get()

        stack = self.exit_stack
        self._set_uris()
        self._set_experiment()

        run_kwargs = get_container(self.config, 'experiment.mlflow.start_run', default={})
        run_kwargs.setdefault('description', self.config.experiment.description)

        run_name = config.experiment.name
        job_id = hydra_config.job.get('id')
        if job_id:
            run_name = f'{run_name}-{job_id}'
        run_kwargs['run_name'] = run_name

        for key, value in sorted(run_kwargs.items()):
            LOGGER.debug('MLflow start_run parameter: %s = %s', key, value)

        active_run = stack.enter_context(mlflow.start_run(**run_kwargs))
        self.active_run_id = active_run.info.run_id

        LOGGER.info(
            'MLflow artifact URI: %s',
            mlflow.get_artifact_uri()
        )

        self.log_parameters()
        return self

    def __exit__(
        self,
        typ: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> bool:
        self.exit_stack.close()
        self.active_run_id = None



#!/usr/bin/env python3

'''
File functions.
'''

import contextlib
import io
import logging
import pathlib
import shutil
import time
from typing import Any


LOGGER = logging.getLogger(__name__)


def backup_copy(path: pathlib.Path) -> None:
    '''
    Attempt to backup a back to a timestamped path in the same directory.

    Args:
        path:
            The path to back up. If it does not exist then nothing will be done.
    '''
    if path.exists():
        while True:
            timestamp = time.strftime('%Y_%m_%d_%H_%M_%S')
            new_suffix = f'.{timestamp}{path.suffix}'
            backup_path = path.with_suffix(new_suffix)
            if not backup_path.exists():
                break
        LOGGER.info('backing up %s to %s', path, backup_path)
        if path.is_dir():
            shutil.copytree(path, backup_path)
        else:
            shutil.copy2(path, backup_path)


@contextlib.contextmanager
def safe_open(path: pathlib.Path, *args: Any, backup: bool = False, **kwargs: Any) -> io.TextIOBase:
    '''
    Context manager for safely opening a path by first creating any missing
    parent directories and then moving any existing file to a backup path.

    Args:
        path:
            The path to open.

        backup:
            If True, backup the file at the path if it exists.

        *args, **kwargs:
            Positional and keyword arguments passed through to
            pathlib.Path.open.
    Returns:
        The open file handle.
    '''
    path = pathlib.Path(path).resolve()
    LOGGER.debug('opening %s', path)

    path.parent.mkdir(parents=True, exist_ok=True)

    if backup:
        backup_copy(path)

    kwargs.setdefault('encoding', 'utf-8')
    with path.open(*args, **kwargs) as handle:  # pylint: disable=unspecified-encoding
        yield handle



#!/usr/bin/env python3
'''
Type definitions and functions.
'''

import pathlib
from typing import Callable, Tuple, Union

from omegaconf import DictConfig


# Generic number type.
Number = Union[int, float]

# Optimization value for the Optuna sweeper returned by hydronaut.run.Runner.__call__.
OptimizationValue = Union[Number, Tuple[Number, ...]]

# Generic path type for functions that can handle both pathlib Paths or strings.
Path = Union[str, pathlib.Path]

# A type supported by the decorator.
Decorable = Union[
    Callable[DictConfig, OptimizationValue],
    Callable[DictConfig, Callable[[], OptimizationValue]]
]

# The return type of the decorator.
Runable = Callable[[], OptimizationValue]



#!/usr/bin/env python3
'''
Structured configs for Hydra and associated helper functions.
'''

import logging
import os
import pathlib
import platform
from dataclasses import dataclass, field
from typing import Optional, Any

from omegaconf import DictConfig, OmegaConf
import hydra
from hydra.utils import to_absolute_path
from hydra.types import RunMode
from hydra.core.config_store import ConfigStore
from hydra.core.global_hydra import GlobalHydra
from hydra.core.hydra_config import HydraConfig
import mlflow

from hydronaut.hydra.resolvers import reregister


HYDRA_VERSION_BASE = '1.2'
LOGGER = logging.getLogger(__name__)

# Per-run paths, used internally to recover current filepaths.
_RUN_PATHS = {}


# This is not supported so use Any.
# DefaultsType = list[dict[str, str] | str]
DefaultsType = list[Any]


# In Python 3.10+ this can be a static method within HydronautConfig but for
# earlier versions this must be scoped outside of the class definition for the
# call to field().
def _get_hydronaut_defaults():
    '''
    Factory for defaults list.
    '''
    return [
        {'override /hydra/job_logging@_global_.hydra.job_logging': 'colorlog'},
        {'override /hydra/hydra_logging@_global_.hydra.hydra_logging': 'colorlog'}
    ]


@dataclass
class MLflowRunConfig:
    '''
    Parameters for mlflow.start_run(), except for experiment_id which is set
    from the experiment name. For details, see
    https://mlflow.org/docs/latest/python_api/mlflow.html
    '''
    #  experiment_id: Optional[str] = None
    run_id: Optional[str] = None
    run_name: Optional[str] = None
    nested: bool = False
    tags: Optional[dict[str, str]] = None
    description: Optional[str] = '${experiment.description}-run_${hydra:job.name}_${hydra:job.num}'


@dataclass
class MLflowConfig:
    '''
    MLflow configuration.
    '''
    start_run: Optional[MLflowRunConfig] = None


@dataclass
class PythonConfig:
    '''
    Pyton configuration.
    '''
    # Paths to prepend to the system path list.
    paths: Optional[list[str]] = field(default_factory=list)


@dataclass
class ExperimentConfig:  # pylint: disable=too-many-instance-attributes
    '''
    Experiment configuration.
    '''
    name: str
    description: str
    params: Optional[dict[str, Any]]
    exp_class: Optional[str] = None
    python: Optional[PythonConfig] = None
    mlflow: Optional[MLflowConfig] = None
    environment: Optional[dict[str, str]] = None
    defaults: Optional[DefaultsType] = field(default_factory=_get_hydronaut_defaults)


def _get_hydronaut_defaults():
    '''
    Factory for defaults list.
    '''
    return ['experiment/hydronaut_experiment']


@dataclass
class HydronautConfig:
    '''
    Common Hydronaut configuration.
    '''
    defaults: Optional[DefaultsType] = field(default_factory=_get_hydronaut_defaults)
    #  experiment: ExperimentConfig


def _get_unique_identifier() -> str:
    '''
    Get a unique process identifier.

    Returns:
        A string with the format "<node>:<pid>".
    '''
    return f'{platform.node()}:{os.getpid()}'


def _reinitialize_from_environment() -> DictConfig:
    '''
    Re-initialize the Hydra configuration in a subprocess from environment
    variables if uninitialized.

    Returns:
        A Hydra config object.
    '''
    LOGGER.debug('Attempting to re-initialize Hydra from environment variables.')

    if GlobalHydra().is_initialized():
        LOGGER.debug('Aborting re-initialization of Hydra: already initialized.')
        return

    my_uid = _get_unique_identifier()
    main_uid = os.getenv('HYDRONAUT_MAIN_UID')
    if my_uid == main_uid:
        LOGGER.debug('Aborting re-initialization of Hydra: current process is the main process.')
        return

    working_dir = os.getenv('HYDRONAUT_WORKING_DIR')
    os.chdir(working_dir)

    GlobalHydra.instance().clear()
    hydra.initialize_config_dir(
        version_base=HYDRA_VERSION_BASE,
        config_dir=os.getenv('HYDRONAUT_CONFIG_DIR'),
        job_name=os.getenv('HYDRONAUT_JOB_NAME')
    )
    cfg = hydra.compose(config_name='config', return_hydra_config=True)
    HydraConfig().set_config(cfg)


def configure_hydra(from_env: bool = False) -> None:
    '''
    Configure the config store, resolvers and global configs as necessary.

    Args:
        from_env:
            If True, re-initialize the Hydra configuration object from
            environment variables set by configure_environment().
    '''
    store = ConfigStore.instance()
    store.store(name='hydronaut_experiment', group='experiment', node=ExperimentConfig)
    store.store(name='hydronaut_config', node=HydronautConfig)
    reregister()
    if from_env:
        _reinitialize_from_environment()


def _safe_resolve_hydra_conf(hydra_conf: DictConfig, conf: DictConfig) -> DictConfig:
    '''
    Safely resolve a Hydra configuration object. This is required for simple
    runs due to a bug in Hydra's default configuration file.

    Args:
        hydra_conf:
            The Hydra configuration object.

        conf:
            The main configuration object, required for resolving interpolations.

    Returns:
        The resolved configuration object except unresolvable interpolations
        will be left in place.
    '''
    # Create a copy to ensure that the original is not modified.
    conf = OmegaConf.create(OmegaConf.to_container(conf, resolve=False))

    # Replace the erroneous "hydra." interpolations with the "hydra:" resolver.
    text = OmegaConf.to_yaml(hydra_conf)
    text = text.replace('${hydra.', '${hydra:')
    hydra_conf = OmegaConf.create(text)

    # Integrate the configuration objects to ensure full resolution.
    conf.hydra = hydra_conf
    OmegaConf.resolve(conf)

    # Extract the Hydra configuration and return it.
    hydra_conf = conf.hydra
    OmegaConf.set_readonly(hydra_conf, True)
    return hydra_conf


def save_config(
    config: OmegaConf,
    path: pathlib.Path,
    resolve: bool = False,
    overwrite: bool = False
) -> None:
    '''
    Save a config to a path.

    Args:
        config:
            An OmegaConf configuration object.

        path:
            The output path.

        resolve:
            Resolve all resolvers before saving the config.
    '''
    if path.exists() and not overwrite:
        LOGGER.warning('Refusing to overwrite existing file: %s', path)
    LOGGER.debug('Saving config to %s', path)
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open('w', encoding='utf-8') as handle:
        OmegaConf.save(config=config, f=handle, resolve=resolve)
    # Make the file read-only to reduce the risk of unintentional changes if the
    # user retrieves the path via get_config_path().
    mode = path.stat().st_mode
    mode &= ~0o333
    path.chmod(mode)
    mlflow.log_artifact(path, artifact_path='hydronaut')


def configure_environment() -> None:
    '''
    Save the current Hydra configuration to fully resolved OmegaConf files and
    set environment variables for subprocesses to re-initialize the current
    configuration.
    '''
    if not HydraConfig.initialized():
        LOGGER.error('Unable to configure environment: Hydra is not initalized')
        return

    hydra_conf = HydraConfig.get()
    job = hydra_conf.job
    if hydra_conf.mode == RunMode.MULTIRUN:
        config_dir = pathlib.Path(hydra_conf.sweep.dir) / str(job.num)
    else:
        config_dir = hydra_conf.run.dir
    config_dir = pathlib.Path(to_absolute_path(config_dir)).resolve()

    hydra_dir = config_dir / hydra_conf.output_subdir
    conf = OmegaConf.load(hydra_dir / 'config.yaml')

    # Configure MLflow
    if conf.experiment.get('mlflow') is None:
        conf.experiment.mlflow = OmegaConf.create(MLflowConfig())

    active_run = mlflow.active_run()
    if active_run is not None:
        os.environ['MLFLOW_RUN_ID'] = str(active_run.info.run_id)
    else:
        LOGGER.error('No active MLflow run')

    # Save resolved versions of the current configs for subprocesses.
    config_dir /= '.hydronaut'
    config_dir.mkdir(parents=True, exist_ok=True)
    hydra_path = config_dir / 'hydra.yaml'
    save_config(
        _safe_resolve_hydra_conf(hydra_conf, conf),
        hydra_path,
        resolve=False,
        overwrite=False
    )
    config_path = config_dir / 'config.yaml'
    save_config(
        conf,
        config_path,
        resolve=True,
        overwrite=False
    )
    # Save the paths so that they can be retrieved via get_config_path().
    _RUN_PATHS[active_run.info.run_id] = {
        'hydra': hydra_path,
        'config': config_path
    }
    #  save_config(GlobalHydra().get(), config_dir / 'config.yaml', resolve=True)

    # Log Hydra configuration files.
    for hydra_config in (config_dir.parent / hydra_conf.output_subdir).glob('*'):
        mlflow.log_artifact(hydra_config, artifact_path='hydra')

    os.environ['HYDRONAUT_CONFIG_DIR'] = str(config_dir)
    os.environ['HYDRONAUT_JOB_NAME'] = str(job.name)
    os.environ['HYDRONAUT_WORKING_DIR'] = str(pathlib.Path.cwd())
    main_uid = os.getenv('HYDRONAUT_MAIN_UID')
    if main_uid is None:
        os.environ['HYDRONAUT_MAIN_UID'] = _get_unique_identifier()


def get_config_path(get_hydra: bool = False) -> pathlib.Path:
    '''
    Get the path to this run's resolved configuration file. This can be useful
    for running external code that cannot be invoked directly with the
    configuration file object.

    Args:
        get_hydra:
            If True, return the path to the configuration file with the
            Hydra-specific configuration. If False (the default), return the
            experiment configuration file path.

    Returns:
        The pathlib.Path object pointing to the requested configuration file, or
        None if the path is not available. The configuration file is fully
        resolved prior to saving and can therefore be used as a normal YAML
        input file.
    '''
    active_run = mlflow.active_run()
    if active_run is None:
        LOGGER.error('Unable to retrieve configuration file path. No active MLflow run.')
        return None
    run_id = active_run.info.run_id
    paths = _RUN_PATHS.get(run_id)
    if not paths:
        LOGGER.error(
            'No configuration paths set for current MLflow run (%s) '
            'configure_environment() should be called first.',
            run_id
        )
        return None
    return paths['hydra' if get_hydra else 'config']



#!/usr/bin/env python3
'''
Hydra-configured convolutional network for MNIST.
'''

import logging

import torch
import pytorch_lightning as pl
from torch.nn import functional as F


from hydronaut.hydra.info import get_sweep_number
from hydronaut.hydra.omegaconf import get_container


LOGGER = logging.getLogger(__name__)


class MNISTClassifier(pl.LightningModule):  # pylint: disable=too-many-ancestors
    '''
    Model generator based on an Optuna Trial.
    '''

    def _add_convolutional_layers(self):
        '''
        Add the sequential convolution layers.

        Returns:
            The output dimension.
        '''
        params = self.config.experiment.params
        input_dim = params.input_dim

        # Variable number of convolutional layers with pooling and drop-out.
        layers = []
        n_layers = params.conv_layers
        current_depth = 1
        size = input_dim

        for i in range(1, n_layers + 1):
            conv_depth = params.get(f'conv_depth_{i}')
            conv_dim = params.get(f'conv_dim_{i}')
            pool = params.get(f'pool_{i}')
            drop = params.get(f'drop_{i}')

            # Size check to ensure minimal dimensions after convolution and
            # pooling.
            size = (size + pool - conv_dim) // pool
            if size <= 1:
                LOGGER.warning(
                    'requested dimensions too small in sweep %d',
                    get_sweep_number()
                )
                break

            layers.extend((
                torch.nn.Conv2d(current_depth, conv_depth, (conv_dim, conv_dim)),
                torch.nn.ReLU(),
                torch.nn.MaxPool2d(pool),
                torch.nn.Dropout(p=drop)
            ))
            current_depth = conv_depth
        layers.append(torch.nn.Flatten(start_dim=-3))
        self.conv_layers = torch.nn.Sequential(*layers)

        # Calculate the number of elements after the convolution layers.
        with torch.no_grad():
            zeros = torch.zeros(  # pylint: disable=no-member
                1, 1, input_dim, input_dim, device=self.device
            )
            return self.conv_layers.forward(zeros).numel()

    def __init__(self, config):
        '''
        Args:
            config:
                Omegaconf configuration object.
        '''
        super().__init__()
        self.config = config
        params = config.experiment.params

        lin1_in = self._add_convolutional_layers()
        lin1_out = params.lin1_out
        lin1_drop = params.lin1_drop

        self.lin1 = torch.nn.Linear(lin1_in, lin1_out)
        # Per-input bias.
        self.register_parameter(
            name='lin1_bias',
            param=torch.nn.Parameter(
                # pylint: disable=no-member
                torch.zeros(lin1_in, requires_grad=True, device=self.device)
            )
        )
        self.lin1_act = torch.nn.ReLU()
        self.lin1_drop = torch.nn.Dropout(p=lin1_drop)

        self.lin2 = torch.nn.Linear(lin1_out, 10)
        self.lin2_act = torch.nn.Softmax(dim=-1)

    def forward(self, x):  # pylint: disable=invalid-name,arguments-differ
        '''
        Parent override.
        '''
        x = self.conv_layers(x)
        x -= self.lin1_bias
        x = self.lin1(x)
        x = self.lin1_act(x)
        x = self.lin1_drop(x)
        x = self.lin2(x)
        x = self.lin2_act(x)
        return x

    def configure_optimizers(self):
        '''
        Parent override.
        '''
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=float(self.config.experiment.params.learning_rate)
        )
        return optimizer

    def cross_entropy_loss(self, logits, labels):
        '''
        Loss function.
        '''
        return F.nll_loss(logits, labels)

    def log_with_config_args(self, *args, **kwargs):
        '''
        Wrapper around log() that auto-appends keyword arguments from the
        config.
        '''
        more_kwargs = get_container(self.config, 'experiment.params.log', default={})
        if more_kwargs:
            kwargs.update(more_kwargs)
        super().log(*args, **kwargs)

    def training_step(self, batch, _batch_idx):  # pylint: disable=arguments-differ
        '''
        Parent override.
        '''
        x, y = batch  # pylint: disable=invalid-name
        logits = self.forward(x)
        loss = self.cross_entropy_loss(logits, y)
        self.log('train_loss', loss.item())
        return loss

    def validation_step(self, batch, _batch_idx):  # pylint: disable=arguments-differ
        '''
        Parent override.
        '''
        x, y = batch  # pylint: disable=invalid-name
        logits = self.forward(x)
        loss = self.cross_entropy_loss(logits, y)
        self.log_with_config_args('val_loss', loss.item())
        return loss

    def test_step(self, batch, _batch_idx):  # pylint: disable=arguments-differ
        '''
        Parent override.
        '''
        x, y = batch  # pylint: disable=invalid-name
        logits = self.forward(x)
        loss = self.cross_entropy_loss(logits, y)
        self.log_with_config_args('test_loss', loss.item())
        return loss



#!/usr/bin/env python3
'''
Decorators for running code with Hydronaut.
'''

import logging
import types
from typing import Callable

from hydronaut.run import Runner
from hydronaut.types import Path, Decorable, Runable


LOGGER = logging.getLogger(__name__)


def with_hydronaut(config_path: Path = None) -> Callable[Decorable, Runable]:
    '''
    A decorator to run custom user code with Hydronaut. The user code must be
    either a class with the following methods:
        __init__(self, config):
            An initialization method that accepts the Hydra configuration object
            as its sole argument.

        __call__(self):
            A call method that returns the experiment's objective values.

        setup(self):
            An optional method that will be invoked before __call__ when the
            experiment is run. This can be used to e.g. prepare data.

    or a function that accept the Hydra configuration object as its sole
    argument and which returns the experiment's objective values.

    Args:
        config_path:
            Passed through to hydronaut.run.Runner.

    Returns:
        A callable object that runs the user code within the Hydronaut framework.
    '''

    # This defines a nested decorator in order to accept the configuration file
    # path as an argument. The first decorator accepts the path and returns the
    # configured decorator that is then used to decorate the user-supplied class
    # or function.

    def _with_hydronaut(user_code: Decorable) -> Runable:
        '''
        Internal decorator.
        '''
        class CustomRunner(Runner):  # pylint: disable=too-few-public-methods
            '''
            Create a custom Runner subclass to wrap the user code.
            '''
            def get_experiment_object(self):
                # If the user code is a simple function, defer calling.
                if isinstance(user_code, types.FunctionType):
                    def exp_obj():
                        return user_code(self.config)
                # Else assume that it is a class.
                else:
                    exp_obj = user_code(self.config)
                    if not isinstance(exp_obj, Callable):
                        LOGGER.warning('%s does not create callable objects.', user_code)

                return exp_obj

        return CustomRunner(config_path=config_path)
    return _with_hydronaut



#!/usr/bin/env python3
'''
Hydra resolvers.
'''

import logging
import multiprocessing
import pathlib
import urllib.parse

try:
    import torch
except ImportError:
    torch = None
from omegaconf import OmegaConf
from hydra.core.utils import setup_globals as hydra_setup_globals
from hydra.utils import get_original_cwd


LOGGER = logging.getLogger(__name__)


def _parse_args(args):
    '''
    Parse resolver arguments.

    Args:
        args:
            A list of strings with the format "key=value" or just "key". Keys
            without a value will be assigned the value of True. If a value is
            given, the words "true", "false", "none" and "null" will be
            converted to the Python values True, False, None and None,
            respectively, regardless of case.

    Returns:
        A dict mapping the keys to the parsed values.
    '''
    kwargs = {}
    val_map = {
        'true': True,
        'false': False,
        'none': None,
        'null': None
    }
    for arg in args:
        try:
            key, value = arg.split('=', 1)
        except ValueError:
            kwargs[arg] = True
        else:
            kwargs[key] = val_map.get(value.lower(), value)
    return kwargs


def get_cwd(*args):
    '''
    The current working directory from which the main function was launched.

    Args:
        resolve:
            If True, resolve symlinks in the path.

        uri:
            If True, return the path as a URI.

    Returns:
        The resulting string.
    '''
    try:
        path = pathlib.Path(get_original_cwd())
    except ValueError:
        path = pathlib.Path.cwd()
    kwargs = _parse_args(args)

    if kwargs.get('resolve', False):
        path = path.resolve()

    if kwargs.get('uri', False):
        return path.as_uri()

    return str(path)


def _get_divisor(args):
    '''
    Internal function to handle optional divisor to some resolvers.
    '''
    if len(args) > 0:
        divisor = int(args[0])
        if divisor > 0:
            return divisor
    return 1


def get_n_cpu(*args):
    '''
    Get the number of available CPUs. Optionally accept an integer argument
    which will be used to divide the result.

    Args:
        *args:
            An optional single integer argument. If given, the number of CPUs
            will be divided by this number.

    Returns:
        The number of available CPUs. If a divisior is given then this will
        always return a value of at least 1.
    '''
    n_cpu = multiprocessing.cpu_count()
    return max(n_cpu // _get_divisor(args), 1)


def get_n_gpu_pytorch(*args):
    '''
    Get the number of available PyTorch GPUs. Optionally accept an integer
    argument which will be used to divide the result.

    Args:
        *args:
            An optional single integer argument. If given, the number of CPUs
            will be divided by this number.

    Returns:
        The number of available PyTorch GPUs. If the number is greater than 0
        than a value of at least 1 will always be returned when an optional
        divisor is given.
    '''
    if torch is None:
        LOGGER.warning('PyTorch must be installed to get the number of PyTorch GPUs')
        return 0
    n_gpu = torch.cuda.device_count()
    if n_gpu < 1:
        return 0
    return max(n_gpu // _get_divisor(args), 1)


def register() -> dict[str, str]:
    '''
    Register custom resolvers.

    Returns:
        A dict mapping resolver names to their descriptions.
    '''

    resolvers = {}
    for name, func in (
            ('url_quote', urllib.parse.quote),
            ('cwd', get_cwd),
            ('n_cpu', get_n_cpu),
            ('n_gpu_pytorch', get_n_gpu_pytorch),
            ('max', max),
            ('min', min)
    ):
        LOGGER.debug('Registing resolver %s', name)
        OmegaConf.register_new_resolver(name, func)
    return resolvers


def reregister() -> None:
    '''
    Reregister custom and Hydra resolvers. This is necessary due to a bug in the
    process launcher.
    '''
    if not OmegaConf.has_resolver('n_cpu'):
        register()
    if not OmegaConf.has_resolver('hydra'):
        hydra_setup_globals()



#!/usr/bin/env python3

'''
Initialize a directory with default files.
'''

import argparse
import logging

from hydronaut.file import safe_open
from hydronaut.log import configure_logging
from hydronaut.paths import PathManager

LOGGER = logging.getLogger()


class Initializer():
    '''
    Initialize a directory with default files.
    '''
    def __init__(self) -> None:
        self.path_manager = PathManager()
        self.config_path = self.path_manager.get_config_path()
        self.pargs = None

    def parse_args(self, args: list[str] = None) -> None:
        '''
        Parse the command-line arguments with argparse.

        Args:
            args:
                The list of arguments to parse. If None, the arguments are taken
                from the current system invocation.

        Returns:
            The parsed arguments.
        '''
        parser = argparse.ArgumentParser(
            description='Generate default files for a Hydronaut project.'
        )
        parser.add_argument(
            '-s', '--sweeper',
            choices=('optuna',),
            help=('''
                Use the given Hydra sweeper.
            ''')
        )
        parser.add_argument(
            '-l', '--launcher',
            choices=('joblib',),
            help=('''
                Use the given Hydra launcher.
            ''')
        )
        self.pargs = parser.parse_args(args)

    def run(self, args: list[str] = None) -> None:
        '''
        Process command-line arguments and generate the requested files.

        Args:
            args:
                Passed through to parse_args().
        '''
        self.parse_args(args=args)
        self.create_config()
        self.create_experiment()

    def create_config(self) -> None:
        '''
        Create the configuration file.
        '''
        pargs = self.pargs
        config_path = self.config_path
        LOGGER.info('creating %s', config_path)
        with safe_open(config_path, 'w', backup=True) as handle:
            handle.write('''
# This is the main Hydra configuration file for the current Hydronaut project.
# Hydra uses OmegaConf to load its configuration files so all OmegaConf features
# are supported. For details, read the respective documentation.
#
# https://hydra.cc/docs/intro/
# https://pypi.org/project/omegaconf/

# Default values are used to override sections of the configuration file with
# values from other files.
defaults:
  # Ensure type-checking for the experiment section. This is specific to
  # Hydronaut. This will also set some defaults for logging configuration.
  - hydronaut_config'''.strip())

            if pargs.sweeper == 'optuna':
                handle.write('''

  # Use the Optuna sweeper. This is configured in the hydra.sweeper section
  # below.
  - override hydra/sweeper: optuna
  # Set the default sampler to TPE. For other supported samplers, see
  # https://optuna.readthedocs.io/en/stable/reference/samplers/index.html
  - override hydra/sweeper/sampler: tpe''')

            if pargs.launcher == 'joblib':
                handle.write('''

  # Use the joblib launcher. This is configured in the hydra.launcher section.
  # https://hydra.cc/docs/1.2/plugins/joblib_launcher/
  - override hydra/launcher: joblib''')

            handle.write('''

  # Optionally enabled colors in log messages.
  - override hydra/job_logging: colorlog
  - override hydra/hydra_logging: colorlog

  # Required to ensure that some of the above settings work correctly.
  - _self_

# Hydra configuration.
hydra:
  # Uncomment this to enable the Hydra multirun with the sweeper by default.
  # Without this option, it is necessary to pass the --multirun option at
  # invocation.
  # mode: MULTIRUN

  # Hydra job configuration: https://hydra.cc/docs/configure_hydra/job/
  job:
    # Set environment variables such as OMP_NUM_THREADS. These variables are set
    # later than those in the "experiment.environment" field below.
    env_set:
        # OMP_NUM_THREADS: ${n_cpu:}
''')
            handle.write('''
  # Sweeper parameter. These depend on the currently configured sweeper.
  sweeper:''')

            if pargs.sweeper == 'optuna':
                handle.write('''
    # Optuna sweeper parameters.
    # See https://hydra.cc/docs/plugins/optuna_sweeper/ for details.

    # The way to optimize the objective function returned by the Experiment
    # subclass (maximize or minimize). This may also be a list if the objective
    # function returns multiple values.
    direction: maximize

    # Set the study name here. The default simply appends "-study" to the
    # experiment's name.
    study_name: ${experiment.name}-study

    # The study storage backend as an SQLAlchemy storage backend URL:
    # https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls
    # If unset or null then the study will use impersistant in-memory storage.
    # storage: sqlite:///${url_quote:"${cwd:}"}/optuna.db


    # The number of Optuna trials.
    n_trials: 40

    # The number of jobs to use. For naive single-CPU models use ${n_cpu:}. For
    # other models such as those based on PyTorch-lightning, set this to 1.
    n_jobs: ${n_cpu:}

    # Sampler options.
    sampler:
      # Set the seed for reproducibility.
      seed: 123''')

            handle.write('''
    # Parameters to sweep. These should append or override values in
    # experiment.params. For details of the available sweeper options such as
    # range, interval and choice, consult the following documentation:
    # https://hydra.cc/docs/plugins/optuna_sweeper/
    params:
      # Add a parameter that will be swept over the integer range [1,10]. This
      # will be accessible from Python via
      # config.experiment.params.swept_param_1.
      ++experiment.params.swept_param_1: range(1, 10)''')

            if pargs.launcher == 'joblib':
                handle.write('''
  # See https://hydra.cc/docs/plugins/joblib_launcher/ for details and further
  # options.
  launcher:
    # Use as many jobs as there are CPUs. This is essentially equivalent to
    # ${n_cpu:}
    n_jobs: -1''')


            handle.write('''

experiment:
  # Experiment name.
  name: My Experiment

  # Experiment description.
  description: Placeholder experiment.

  # The Experiment subclass to use for the experiment. This is a string with the
  # format "<module>:<class name>". "<module>" must be the name of a module that
  # Python can import and it must contain a subclass of
  # hydronaut.experiment.Experiment with the name "<class name>". The default is
  # set to the class in the generated in the src/experiment.py file.
  exp_class: experiment:MyExperiment

  # Experiment parameters used in the Experiment subclass to configure the
  # model, data and everything else. If a sweeper is used, then the values will
  # be appended to this section before the config object is passed to the
  # Experiment subclass.
  params:
    # Example parameter that will be accessible from Python via
    # config.experiment.params.fixed_param_1.
    fixed_param_1: 5

  # Python configuration.
  python:
    # System search paths relative to the working directory. Use this to add
    # custom modules to the Python system path without having to manually
    # configure PYTHONPATH or sys.path.
    paths:
        - src

  # Environment variables. This section may be used to configure MLflow
  # tracking, artifact and registry servers, for example. It should map
  # environment variable names to their values, as strings.
  # For example, see
  # https://mlflow.org/docs/latest/tracking.html
  # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html#environment-variables
  environment:
    # Examples of common variables:
    # MLFLOW_TRACKING_URI:
    # MLFLOW_TRACKING_USERNAME:
    # MLFLOW_TRACKING_PASSWORD:
    # MLFLOW_S3_ENDPOINT_URL:
    # AWS_ACCESS_KEY_ID:
    # AWS_SECRET_ACCESS_KEY:
    # AWS_DEFAULT_REGION:
    # AWS_SESSION_TOKEN:

  # MLflow configuration
  mlflow:
    # Run parameters. These are passed through directly to
    # mlflow.start_run().See the documentation for details:
    # https://mlflow.org/docs/latest/python_api/mlflow.html
    start_run:
      # Set a description for the run. This defaults to the value of the
      # "experiment.description" field above. It can be customized here to
      # include run-specific parameters, for example.
      description: ${experiment.description}
      # Set optional arbitrary tags. These could contain parameter values, dates
      # and times, the hostname of the machine on which the run was executed,
      # the name of the user who executed the run, etc.
      tags:
        framework: Hydronaut
''')

    def create_experiment(self) -> None:
        '''
        Create a placeholder experiment subclass.
        '''
        source_path = self.path_manager.get_src_path('experiment.py')
        LOGGER.info('creating %s', source_path)
        with safe_open(source_path, 'w', backup=True) as handle:
            handle.write("""#!/usr/bin/env python3
'''
Placeholder experiment.
'''

from hydronaut.experiment import Experiment


class MyExperiment(Experiment):
    '''
    Placeholder experiment to serve as a starting point.
    '''

    # This is the only required definition to use the Hydronaut framework. The
    # __call__ method should encapsulate your workflow from data loading to
    # model fitting and then return one or more metrics for the optimizer.
    #
    # All parameters should be set in the config object that Hydra creates from
    # your YAML configuration file (conf/config.yaml).
    def __call__(self):
        # The configuration object.
        config = self.config
        # The parameters set in the experiment.params section.
        params = config.experiment.params

        # Use the configured parameters.
        fixed_param_1 = params.fixed_param_1
        # "get()" is used here because the parameter will only be set when Hydra
        # is in the multirun mode.
        swept_param_1 = params.get('swept_param_1', 1)

        # Return a value.
        return fixed_param_1 * swept_param_1""")


def main(args: list[str] = None) -> None:
    '''
    Initialize a directory with default files.
    '''
    configure_logging()
    initializer = Initializer()
    initializer.run(args=args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
'''
Omegaconf utility functions.
'''

import importlib
import inspect
import logging
from typing import Any, Optional

from omegaconf import OmegaConf, DictConfig  # pylint: disable=import-self


LOGGER = logging.getLogger(__name__)


def get(config: DictConfig, name: str, default: Any = None) -> Any:
    '''
    Get a nested parameter such as foo.bar.baz from an Omegaconf configuration
    object, with default value.

    Args:
        config:
            The configuration object.

        name:
            The nested parameter name. It will be split on ".".

        default:
            An optional default value. Default: None

    Returns:
        The parameter value, or the default value if any of the nested values
        are missing or the resulting value is None.
    '''
    value = config
    for component in name.split('.'):
        try:
            value = getattr(value, component)
        except AttributeError:
            return default
    if value is None:
        return default
    return value


def get_container(
    config: DictConfig,
    name: str,
    *,
    default: Any = None,
    resolve: bool = True
) -> Any:
    '''
    A wrapper around get() that transforms the result with OmegaConf.to_container.

    Args:
        config:
            Same as get().

        name:
            Same as get().

        default:
            Same as get().

        resolve:
            Passed through to to_container(). If True, the resolvers will be
            resolved before creating the container.

    Returns:
        The container returned by a call to OmegaConf.to_conainer if the result
        of get() is an OmegaConf configuration object, otherwise the value
        returned by get().
    '''
    result = get(config, name, default=default)
    if OmegaConf.is_config(result):
        result = OmegaConf.to_container(result, resolve=resolve)
    return result


def set_globals(
    config: DictConfig,
    name: str,
    *,
    module: Optional[str] = None,
    resolve: bool = True
) -> None:
    '''
    Set or update global values in a loaded module with values from the
    configuration object. This is mainly useful for speeding up calculations
    that would otherwise suffer from repeated lookups up the values within the
    OmegaConf configuration object.

    Args:
        config:
            Same as get().

        name:
            Same as get().

        module:
            An optional module or module name. If not given then the module from
            which this function was invoked is used.

        resolve:
            Same as get_container().
    '''
    values = get_container(config, name, default={}, resolve=resolve)
    if not values:
        return
    if module is None:
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
    elif isinstance(module, str):
        module = importlib.import_module(module)
    elif not inspect.ismodule(module):
        raise ValueError(
            f'The value of the module parameter must be a module, a string or None, '
            f'not {type(module)}'
        )
    for key, value in values.items():
        LOGGER.debug('Setting %s.%s to %s', module.__name__, key, value)
        setattr(module, key, value)



#!/usr/bin/env python3
'''
Dummy function with a decorator.
'''

import logging
import sys

from hydronaut.decorator import with_hydronaut


LOGGER = logging.getLogger(__name__)


@with_hydronaut(config_path='conf/config2.yaml')
class Dummy():
    '''
    A dummy class to show the user of the Hydronaut decorator. It is based on
    the dummy Experiment subclass example.
    '''
    def __init__(self, config):
        '''
        Args:
            config:
                The configuration object which will be passed from the decorator
                at runtime.
        '''
        self.config = config

    def setup(self):
        '''
        The setup method will be called before __call__ at runtime. It can be
        used to prepare data for the experiment or perform other actions that
        need only be done once.
        '''
        LOGGER.info('called setup()')

    def __call__(self):
        # Get the experiment parameters from the config object.
        params = self.config.experiment.params

        # Return some numerical value based on the parameters. In a real
        # experiment, this would be a validation or test score after training
        # and fitting a model.
        return params.x * params.y * params.z


if __name__ == '__main__':
    sys.exit(Dummy())  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python3
'''
MNIST experiment.
'''

import logging

from omegaconf import DictConfig

from hydronaut.torch.lightning.experiment import PLExperiment

from model import MNISTClassifier  # pylint: disable=wrong-import-order,no-name-in-module
from data import MNISTDataModule  # pylint: disable=wrong-import-order

LOGGER = logging.getLogger(__name__)


class MNISTExperiment(PLExperiment):  # pylint: disable=too-few-public-methods
    '''
    MNIST experiment with PyTorch Lightning.
    '''

    # The PyTorch Lighting Experiment class only requires the user to define a
    # model and a dataloader.
    def __init__(self, config: DictConfig):
        super().__init__(
            config,
            MNISTClassifier,
            MNISTDataModule
        )

    def get_objective_value(self):
        # Return the value logged by the LightningModule instance in the
        # test_step method.
        return self.results['test'][-1]['test_loss']



#!/usr/bin/env python3
'''
PyTorch Basic MNIST Example Experiment.
'''

import torch

from hydronaut.experiment import Experiment

# Import the main function from the adapted example code.
from model import main  # pylint: disable=wrong-import-order


class MyExperiment(Experiment):  # pylint: disable=too-few-public-methods
    '''
    Invoke the code in mode.py
    '''

    def __call__(self):
        # The configuration object is an attribute of the parent class.
        # Pass it to the main function which was adapted to use it and
        # return the mode, the average loss and the accuracy.
        model, avg_loss, accuracy = main(self.config)

        # Optionally save the model.
        if self.config.experiment.params.meta.save_model:
            torch.save(model.state_dict(), 'mnist_cnn.pt')
            self.log_artifact('mnist_cnn.pt', 'model')

        # Log both metrics
        self.log_metric('average loss', avg_loss)
        self.log_metric('accuracy', accuracy)

        # Return the accuracy as the objective value to maximize.
        # We could just as well minimize the average loss, or even
        # return both values to optimize both simultaneous. We'll see
        # this in a later tutorial.
        return accuracy



#!/usr/bin/env python3
'''
PyTorch Lightning experiments.
'''

import logging

import mlflow.pytorch
import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback
from omegaconf import DictConfig

from hydronaut.experiment import Experiment
from hydronaut.hydra.omegaconf import get_container
from hydronaut.types import OptimizationValue

LOGGER = logging.getLogger(__name__)


class PLExperiment(Experiment):
    '''
    Experiment class for PyTorch Lightning experiments.

    The configuration file should include a dictionary under
    "experiment.params.trainer" that contains PyTorch Lighting Trainer keyword
    arguments and their values. These will be passed through to the Trainer.
    '''

    def __init__(
        self,
        config: DictConfig,
        model_cls: pl.LightningModule,
        data_module_cls: pl.LightningDataModule,
        callback_classes: list[Callback] = None
    ) -> None:
        '''
        Args:
            config:
                Same as Experiment.__init__().

            model_cls:
                A subclass of LightningModule.

            data_module_cls:
                A subclass of LightningDataModule.

            callback_classes:
                An optional list of Callback subclasses.

        model_cls, data_module_cls and all classes in callback_classes should
        all accept a single configuration object as a parameter to their
        __init__ methods.
        '''
        super().__init__(config)
        self.model_cls = model_cls
        self.data = data_module_cls(config)
        if callback_classes:
            self.callbacks = [c_cls(config) for c_cls in callback_classes]
        else:
            self.callbacks = []
        self.trainer = None

        # Results object. Currently it only stores the results of the call to
        # trainer.test.
        self.results = {}

    def get_objective_value(self) -> OptimizationValue:
        '''
        Return an objective value for the optimizer. This must be overridden in
        subclasses. It can use any of the attributes of this class such as the
        LightningModule instance or any of the optional Callback instances.

        Returns:
            A numerical value or tuple of numerical values for the optimizer.
        '''
        raise NotImplementedError(
            f'{self.__class__.__name__} does not override get_objective_value()'
        )

    def setup(self):
        '''
        Invoke the data modules setup function to download the data before the
        trainer creates multiple processes that may lead to file conflicts.
        '''
        self.data.setup()

    def __call__(self):
        config = self.config
        params = config.experiment.params

        model = self.model_cls(config)
        data = self.data
        trainer_kwargs = get_container(params, 'trainer', default={}, resolve=True)
        trainer_kwargs['callbacks'] = self.callbacks

        trainer = pl.Trainer(**trainer_kwargs)
        self.trainer = trainer

        # Configure MLflow logging.
        # https://github.com/mlflow/mlflow/blob/eb3588abb55032838b1ff5af1b7d660f2b826dee/examples/pytorch/MNIST/mnist_autolog_example.py
        if trainer.global_rank == 0:
            mlflow.pytorch.autolog()

        trainer.fit(model, datamodule=data)
        self.results['test'] = trainer.test(datamodule=data)

        return self.get_objective_value()



#!/usr/bin/env python3
'''
Hydra utility functions.
'''

from hydra.core.hydra_config import HydraConfig


def get_sweep_number() -> int:
    '''
    Get the number of the current sweep.

    Returns:
        An int between in the range [0, n-1] where n is the number of sweep
        runs.
    '''
    return HydraConfig.get().job.num



#!/usr/bin/env python3
'''
Experiment base class.
'''

import logging
import pathlib
from typing import Any, Optional

import mlflow
from omegaconf import DictConfig

from hydronaut.types import Number, Path, OptimizationValue


LOGGER = logging.getLogger(__name__)


class Experiment():
    '''
    Base class for experiments.

    TODO
    * Add methods for logging parameters, artefacts, etc.
    '''
    def __init__(self, config: DictConfig) -> None:
        '''
        Args:
            config:
                The Hydra configuration object.
        '''
        self.config = config
        # This is set dynamically by Hydronaut at runtime.
        self.mlflow_runner = None

    def __call__(self) -> OptimizationValue:
        '''
        Run the experiment and return the objective value(s).

        Returns:
            The objective value, or values when optimizing a multi-objective
            parameter space.
        '''
        raise NotImplementedError(f'{self.__class__.__name__} does not implement __call__')

    def setup(self) -> None:
        '''
        Set up the experiment before running it (e.g. download data, generate
        files, configure databases).
        '''

    @staticmethod
    def _get_artifact_uri(path: Path, subdir: Optional[Path] = None) -> str:
        '''
        Get the full URI to an artifact.

        Args:
            path:
                The path to the artifact.

            subdir:
                The subpath under the URI in which the artifact is stored.

        Returns:
            The artifact URI.
        '''

        if subdir is not None:
            path = pathlib.Path(subdir) / path
        return mlflow.get_artifact_uri(str(path))

    def log_artifact(self, local_path: Path, artifact_path: Optional[Path] = None) -> None:
        '''
        Wrapper around mlflow.log_artifact:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_artifact

        Args:
            local_path:
                A path to an existing artifact. The artitfact must already be
                saved to this path.

            artifact_path:
                An optional path to a directory under artifact_uri in which to
                log the artifact in MLflow.
        '''
        uri = self._get_artifact_uri(local_path, subdir=artifact_path)
        local_path = str(local_path)
        if artifact_path is not None:
            artifact_path = str(artifact_path)
        LOGGER.debug('logging artifact %s to %s', local_path, uri)
        mlflow.log_artifact(local_path, artifact_path=artifact_path)

    def log_artifacts(self, local_dir: Path, artifact_path: Optional[Path] = None) -> None:
        '''
        Wrapper around mlflow.log_artifacts:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_artifacts

        Args:
            local_dir:
                A path to a directory with artifacts.

            artifact_path:
                An optional path to a directory under artifact_uri in which to
                log the artifact in MLflow.
        '''
        uri = self._get_artifact_uri(local_dir, subdir=artifact_path)
        local_dir = str(local_dir)
        if artifact_path is not None:
            artifact_path = str(artifact_path)
        LOGGER.debug('logging artifacts in %s to %s', local_dir, uri)
        mlflow.log_artifacts(local_dir, artifact_path=artifact_path)

    def log_dict(self, dictionary: Any, artifact_file: Path) -> None:
        '''
        Wrapper around mlflow.log_dict:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_dict

        Args:
            dictionary:
                A dict object to log.

            artifact_file:
                The run-relative subpath to which to save the data as a JSON or
                YAML file. The file type will be determined from the extension
                (".yaml" or ".json").
        '''
        uri = self._get_artifact_uri(artifact_file)
        artifact_file = str(artifact_file)
        LOGGER.debug('logging dict to %s', uri)
        mlflow.log_dict(dictionary, artifact_file)

    def log_figure(self, figure: Any, artifact_file: Path) -> None:
        '''
        Wrapper around mlflow.log_figure:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_figure

        Args:
            figure:
                A figure to log. Check the documentation of mlflow.log_figure
                for supported figure types (e.g. Matplotlib and Plotly.py
                figures).

            artifact_file:
                The run-relative subpath to which to save the figure.
        '''
        uri = self._get_artifact_uri(artifact_file)
        artifact_file = str(artifact_file)
        LOGGER.debug('logging figure to %s', uri)
        mlflow.log_figure(figure, artifact_file)

    def log_image(self, image: Any, artifact_file: Path) -> None:
        '''
        Wrapper around mlflow.log_image:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_image

        Args:
            image:
                An image to log. Check the documentation of mlflow.log_image for
                supported image types (e.g. numpy.ndarray, Pillow images).

            artifact_file:
                The run-relative subpath to which to save the image.
        '''
        uri = self._get_artifact_uri(artifact_file)
        artifact_file = str(artifact_file)
        LOGGER.debug('logging image to %s', uri)
        mlflow.log_figure(image, artifact_file)

    def log_metric(self, key: str, value: Number, step: Optional[int] = None) -> None:
        '''
        Wrapper around mlflow.log_metric:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_metric

        Args:
            key:
                The metric name.

            value:
                The metric value.

            step:
                The metric step. Defaults to zero if unspecified.
        '''

        LOGGER.debug('logging metric %s = %s [step: %s]', key, value, step)
        mlflow.log_metric(key, value, step=step)

    def log_metrics(self, metrics: dict[str, Number], step: Optional[int] = None) -> None:
        '''
        Wrapper around mlflow.log_metrics:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_metrics

        Args:
            metrics:
                A dict mapping metric names to their values.

            step:
                The metric step. Defaults to zero if unspecified.
        '''

        LOGGER.debug('logging metrics %s [step: %s]', metrics, step)
        mlflow.log_metrics(metrics, step=step)

    def log_param(self, key: str, value: Any) -> None:
        '''
        Wrapper around mlflow.log_param:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_param

        Args:
            key:
                The parameter name.

            value:
                The parameter value.
        '''

        LOGGER.debug('logging parameter %s = %s', key, value)
        mlflow.log_param(key, value)

    def log_params(self, params: dict[str, Any]) -> None:
        '''
        Wrapper around mlflow.log_params:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_params

        Args:
            params:
                A dict mapping parameter names to their values.
        '''

        LOGGER.debug('logging parameters %s', params)
        mlflow.log_params(params)

    def log_text(self, text: str, artifact_file: Path) -> None:
        '''
        Wrapper around mlflow.log_text:
        https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_text

        Args:
            text:
                The text to log.

            artifact_file:
                The run-relative subpath to which to save the text.
        '''
        uri = self._get_artifact_uri(artifact_file)
        artifact_file = str(artifact_file)
        LOGGER.debug('logging text to %s', uri)
        mlflow.log_text(text, artifact_file)



#!/usr/bin/env python3

'''
Adaptation of the PyTorch Basic MNIST Example:
https://github.com/pytorch/examples/blob/main/mnist/main.py
'''
import logging

import torch
from torch import nn
import torch.nn.functional as F
from torch import optim
from torch.optim.lr_scheduler import StepLR
from torchvision import datasets, transforms


# Use a standard logger to replace the print statements for homogeneous output.
# Even without Hydronaut it is recommended to use standard logging functions.
LOGGER = logging.getLogger(__name__)


class Net(nn.Module):  # pylint: disable=too-few-public-methods
    '''
    Example nn.Module subclass.
    '''
    def __init__(self, params):
        super().__init__()

        # Get the model parameters from the configuration object.
        self.params = params
        conv1 = params.conv1
        conv2 = params.conv2

        # Calculate the input size to the fully connected layer based on the current
        # parameters.
        #
        # Each convolution layer reduces each image dimension by ( [(D - K) / S] + 1) where
        # D is the input dimension (width or height), K is the kernel size, [] represents the
        # floor function, and S is the stride.
        #
        # The pooling layer applied below in forward() also reduces the dimension by [D / P]
        # where P is the value passed to max_pool2d.
        #
        # The resulting flattened size is D * D * CO_2 where CO_2 is the number of output
        # channels from the second convolution layer.
        input_dim = 28  # fixed MNIST input dimension
        after_conv1_dim = ((input_dim - conv1.kernel_size) // conv1.stride) + 1
        after_conv2_dim = ((after_conv1_dim - conv2.kernel_size) // conv2.stride) + 1
        after_pool_dim = after_conv2_dim // params.max_pool2d
        fc_size = after_pool_dim * after_pool_dim * conv2.out_channels

        # The updated code.
        self.conv1 = nn.Conv2d(1, conv1.out_channels, conv1.kernel_size, conv1.stride)
        self.conv2 = nn.Conv2d(
            conv1.out_channels,
            conv2.out_channels,
            conv2.kernel_size,
            conv2.stride
        )
        self.dropout1 = nn.Dropout(params.dropout1)
        self.dropout2 = nn.Dropout(params.dropout2)
        self.fc1 = nn.Linear(fc_size, params.fc1)
        self.fc2 = nn.Linear(params.fc1, 10)

    def forward(self, x):  # pylint: disable=invalid-name
        '''
        Forward method.
        '''
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, self.params.max_pool2d)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)  # pylint: disable=no-member
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


def train(  # pylint: disable=too-many-arguments
    params,
    model,
    device,
    train_loader,
    optimizer,
    epoch
):
    '''
    Train the model.

    Args:
        params:
            Training parameters.

        model:
            The model.

        device:
            The PyTorch device on which to train the model.

        train_loader:
            The DataLoader for the training data.

        optimizer:
            The training optimizer.

        epoch:
            The epoch number.
    '''
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % params.log_interval == 0:
            # Transform a print statement to a logging function.
            LOGGER.info(
                'Train Epoch: %s [%d/%d (%.0f%%)]\tLoss: %0.6f',
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), loss.item()
            )
            if params.dry_run:
                break


def test(model, device, test_loader):
    '''
    Test the trained model.

    Args:
        model:
            The trained model.

        device:
            the PyTorch device on which to test the model.

        test_loader:
            The Dataloader for the test data.

    Returns:
        The test loss and accuracy.
    '''
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)
    accuracy = correct / len(test_loader.dataset)

    # Transform a print statement to a logging function.
    LOGGER.info(
        'Test set: Average loss: %0.4f, Accuracy %d/%d (%.0f%%)',
        test_loss, correct, len(test_loader.dataset),
        100. * accuracy
    )

    # Return some metrics for optimization.
    return test_loss, accuracy


def main(config):  # pylint: disable=too-many-locals
    '''
    Load the data then train and test the model.

    Returns:
        The trained model, the average loss and the test accuracy.
    '''
    params = config.experiment.params
    meta_params = params.meta
    training_params = params.training

    use_cuda = not meta_params.no_cuda and torch.cuda.is_available()
    use_mps = not meta_params.no_mps and torch.backends.mps.is_available()

    torch.manual_seed(meta_params.seed)

    if use_cuda:
        device = torch.device("cuda")  # pylint: disable=no-member
    elif use_mps:
        device = torch.device("mps")  # pylint: disable=no-member
    else:
        device = torch.device("cpu")  # pylint: disable=no-member

    train_kwargs = {'batch_size': training_params.batch_size}
    test_kwargs = {'batch_size': training_params.test_batch_size}
    if use_cuda:
        cuda_kwargs = {'num_workers': 1,
                       'pin_memory': True,
                       'shuffle': True}
        train_kwargs.update(cuda_kwargs)
        test_kwargs.update(cuda_kwargs)

    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
        ])
    dataset1 = datasets.MNIST('../data', train=True, download=True, transform=transform)
    dataset2 = datasets.MNIST('../data', train=False, transform=transform)
    train_loader = torch.utils.data.DataLoader(dataset1, **train_kwargs)
    test_loader = torch.utils.data.DataLoader(dataset2, **test_kwargs)

    model = Net(params.model).to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=training_params.lr)

    scheduler = StepLR(optimizer, step_size=1, gamma=training_params.gamma)
    # Get the metrics
    avg_loss = None
    accuracy = None
    for epoch in range(1, training_params.epochs + 1):
        train(meta_params, model, device, train_loader, optimizer, epoch)
        avg_loss, accuracy = test(model, device, test_loader)
        scheduler.step()

    # Return the model and the metrics.
    return model, avg_loss, accuracy



#!/usr/bin/env python3

'''
Path management.
'''

import logging
import os
import pathlib
import sys


from hydronaut.types import Path


LOGGER = logging.getLogger(__name__)


class PathManager():
    '''
    Path manager. This provides methods to retrieve common paths.
    '''

    # The default subpath to the main configuration file. It will be
    # interpreted relative to the working directory.
    DEFAULT_CONFIG_SUBPATH = 'conf/config.yaml'

    # Enviroment variable for changing the main configuration file.
    HYDRONAUT_CONFIG_ENV_VAR = 'HYDRONAUT_CONFIG'

    def __init__(self, base_dir: Path = None, config_path: Path = None) -> None:
        '''
        Args:
            base_dir:
                The directory relative to which to interpret paths. If None, the
                current working directory is used.
        '''
        self.base_dir = pathlib.Path.cwd() if base_dir is None else pathlib.Path(base_dir).resolve()
        if config_path is not None:
            config_path = pathlib.Path(config_path).resolve()
        self._config_path = config_path
        LOGGER.debug('PathManager base directory: %s', self.base_dir)

    def add_python_paths(self, paths: list[Path]) -> None:
        '''
        Add Paths to the Python system paths list. The resulting list will be the
        equivalent of concatenating the input paths with the current system paths.

        Args:
            paths:
                The paths to add. If relative, they are interpreted relative to the
                path given by relative_to.
        '''
        if not paths:
            return
        paths = [str(self.base_dir.joinpath(path)) for path in paths]
        for path in paths:
            LOGGER.debug('adding %s to Python system path', path)
        sys.path[:] = [*paths, *sys.path]

    @property
    def config_path(self) -> pathlib.Path:
        '''
        The resolved path to the main configuration file as a pathlib.Path
        object. If the path was not specified during initialization and the
        environment variable is not set, it returns the default path.
        '''
        if self._config_path is not None:
            return self._config_path
        env_path = os.getenv(self.HYDRONAUT_CONFIG_ENV_VAR)
        if env_path is None:
            return self.base_dir / self.DEFAULT_CONFIG_SUBPATH
        env_path = pathlib.Path(env_path)
        if env_path.is_absolute():
            return env_path.resolve()
        return (self.base_dir / env_path).resolve()

    @property
    def config_dir(self) -> pathlib.Path:
        '''
        The resolved path to the current configuration directory as a
        pathlib.Path object.
        '''
        return self.config_path.parent

    def get_config_path(self, subpath: Path = None) -> pathlib.Path:
        '''
        Get the path to a configuration file.

        subpath:
            The subpath relative to the configuration directory. If None,
            defaults to config.yaml.

        Returns:
            The resolved path to the configuration file as a pathlib.Path
            object.
        '''
        if subpath is None:
            return self.config_path
        return self.config_dir / subpath

    def get_src_path(self, subpath: Path) -> pathlib.Path:
        '''
        Get the path to a Python source file in the default source directory.

        subpath:
            The subpath relative to the source directory.

        Returns:
            The resolved path to the source file as a pathlib.Path object.
        '''
        return self.base_dir / 'src' / subpath



#!/usr/bin/env python3
'''
Run an experiment with Hydra, Optuna and MLflow.
'''

import importlib
import inspect
import logging
import os
import subprocess
import sys
from typing import Any, Callable

import hydra
import mlflow

from hydronaut.experiment import Experiment
from hydronaut.hydra.config import configure_hydra, configure_environment, HYDRA_VERSION_BASE
from hydronaut.hydra.omegaconf import get, get_container
from hydronaut.mlflow import MLflowRunner
from hydronaut.paths import PathManager
from hydronaut.types import Path, OptimizationValue
from hydronaut.version import VERSION

LOGGER = logging.getLogger(__name__)


class Runner():
    '''
    Experiment runner.
    '''

    def __init__(self, config_path: Path = None) -> None:
        '''
        Args:
            config_path:
                The path to the configuration file. It is passed through to
                hydronaut.paths.get_config_path to get the main configuration
                file.
        '''
        self.config_path = config_path
        self._path_manager = PathManager(config_path=config_path)
        # The Hydra configuration object, which will be set when this is run.
        self.config = None
        self.mlflow_runner = None
        self._configure_logging()

    def get_experiment_object(self) -> Experiment:
        '''
        Get the experiment class specified by the configuration object.

        Returns:
            The experiment instance.

        Raises:
            ImportError:
                The module could not be found.

            AttributeError:
                The module does not contain the expected class name.
        '''
        config = self.config
        exp_cls_param = config.experiment.exp_class
        try:
            exp_cls_module, exp_cls_name = exp_cls_param.split(':', 1)
        except (AttributeError, ValueError) as err:
            LOGGER.error(
                'Missing or invalid experiment class specification in %s: '
                'experiment.exp_class is %s',
                self._path_manager.config_path,
                exp_cls_param
            )
            raise err

        importlib.import_module(exp_cls_module)
        exp_cls = getattr(sys.modules[exp_cls_module], exp_cls_name)
        LOGGER.debug('Loaded experiment class: %s', exp_cls)

        if not issubclass(exp_cls, Experiment):
            LOGGER.warning(
                'Experiment class %s.%s [%s] is not a subclass of %s.%s',
                exp_cls.__module__,
                exp_cls.__qualname__,
                inspect.getfile(exp_cls),
                Experiment.__module__,
                Experiment.__qualname__
            )
        return exp_cls(config)

    def _configure_logging(self) -> None:
        '''
        Configure initial logging. Hydra will override this when the run is
        started.
        '''
        # Use the default Hydra format and level.
        logging.basicConfig(
            style='{',
            format='[{asctime:s}][{name:s}][{levelname:s}] - {message:s}',
            level=logging.INFO
        )

    def _configure_environment(self) -> None:
        '''
        Set environment variables defined in the configuration file.
        '''
        env_vars = get_container(self.config, 'experiment.environment', default={})
        for name, value in env_vars.items():
            os.environ[name] = value

    def _log_missing_config(self, config_path):
        LOGGER.error(
            'The expected configuration file (%s) does not exist.',
            config_path
        )
        if self.config_path is None:
            LOGGER.warning(
                'If you wish to use a different file, set the path in '
                'the %s environment variable.',
                self._path_manager.HYDRONAUT_CONFIG_ENV_VAR
            )
        LOGGER.warning(
            'Relative paths are interpreted relative to %s.',
            self._path_manager.base_dir
        )

    def _try_setup(self, obj):
        '''
        Attempt to run the object's "setup" method if it has one. This is
        required because the object may be a decorated user class or function
        without a setup method.
        '''
        # This is not as Pythonic as a try-except block but this avoid
        # accidentally catching exceptions raised within the setup method.
        if hasattr(obj, 'setup'):
            if isinstance(obj.setup, Callable):
                obj.setup()
            else:
                LOGGER.warning('"setup" method of %s is not callable', obj)
                try:
                    obj.setup()
                except AttributeError:
                    pass

    def __call__(self, *args: Any, **kwargs: Any) -> OptimizationValue:
        '''
        Run the experiment and return a value for the Optuna sweeper.

        Args:
            *args, **kwargs:
                Positional and keyword arguments passed through to the
                hydra.main function.

        Returns:
            The value to optimize.
        '''
        configure_hydra()
        config_path = self._path_manager.config_path
        if not config_path.exists():
            self._log_missing_config(config_path)
            return 1

        @hydra.main(
            version_base=HYDRA_VERSION_BASE,
            config_path=str(config_path.parent),
            config_name=config_path.stem
        )
        def _run(config):
            '''
            Internal runner function. This is defined within this method so that
            class attributes can be used as arguments to hydra.main().

            Args:
                config:
                    The Hydra configuration object. It must contain a field
                    named "experiment.class" that points to an importable Python
                    class which is a subclass of
                    hydronaut.experiment.Experiment.
            '''
            LOGGER.info('Hydronaut version: %s', VERSION)
            LOGGER.info('Hydronaut configuration file: %s', config_path)
            configure_hydra()
            self.config = config
            self._configure_environment()

            with MLflowRunner(config, base_dir=self._path_manager.base_dir) as runner:
                self.mlflow_runner = runner
                # Configure the environment variables so that any experiment
                # subprocesses can invoke configure_hydra(from_env=True) to
                # re-establish the Hydra configuration. This also ensures that they
                # set the right MLflow run ID.
                configure_environment()

                self._path_manager.add_python_paths(
                    get(self.config, 'experiment.python.paths')
                )

                exp = self.get_experiment_object()
                exp.mlflow_runner = runner
                self._try_setup(exp)
                obj_val = exp()
                mlflow.log_metric("Objective Value", obj_val)
                return obj_val
        return _run(*args, **kwargs)


def main(*args: Any, **kwargs: Any) -> OptimizationValue:
    '''
    Main function to run an experiment with the Hydra configuration.

    Args:
        *args, **kwargs:
            Positional and keyword arguments passed through to the hydra.main
            function.

    Returns:
        The value returned by calling the runner.
    '''
    return Runner()(*args, **kwargs)


def script_main() -> None:
    '''
    Hydra makes some assumptions about configuration paths based on how the main
    function is called. This is a workaround for creating a script via
    pyproject.toml. It will simply invoke main() with any passed command-line
    arguments.
    '''
    cmd = (sys.executable, '-m', 'hydronaut.run', *sys.argv[1:])
    try:
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as err:
        sys.exit(err)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
'''
Data loaders.
'''

import logging
import pathlib

from torch.utils.data import DataLoader, random_split
from torchvision.datasets import MNIST
from torchvision import transforms
import pytorch_lightning as pl


from hydronaut.hydra.omegaconf import get, get_container
from hydronaut.hydra.config import configure_hydra


LOGGER = logging.getLogger(__name__)


class MNISTDataModule(pl.LightningDataModule):
    '''
    MNIST data loader.
    '''
    def __init__(self, config):
        '''
        Args:
            config:
                Omegaconf configuration object.
        '''
        super().__init__()
        self.config = config
        self.mnist_train = None
        self.mnist_val = None
        self.mnist_test = None

        self.dataloader_kwargs = get_container(
            self.config,
            'experiment.params.dataloader',
            default={}
        )

    def setup(self, stage=None):
        '''
        Setup method.
        '''
        # Required when the worker is run in a thread.
        configure_hydra(from_env=True)
        data_dir = str(
            pathlib.Path(
                get(
                    self.config,
                    'experiment.params.data_directory',
                    default='tmp'
                )
            ).resolve()
        )

        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ])

        mnist_train = MNIST(data_dir, train=True, download=True, transform=transform)
        self.mnist_test = MNIST(data_dir, train=False, download=True, transform=transform)

        self.mnist_train, self.mnist_val = random_split(mnist_train, [55000, 5000])

    def train_dataloader(self):
        '''
        train_dataloader method.
        '''
        return DataLoader(self.mnist_train, **self.dataloader_kwargs)

    def val_dataloader(self):
        '''
        val_dataloader method.
        '''
        return DataLoader(self.mnist_val, **self.dataloader_kwargs)

    def test_dataloader(self):
        '''
        test_dataloader method.
        '''
        return DataLoader(self.mnist_test, **self.dataloader_kwargs)



#!/usr/bin/env python3
'''
Dummy function with a decorator.
'''

import sys

from hydronaut.decorator import with_hydronaut


@with_hydronaut(config_path='conf/config.yaml')
def main(config):
    '''
    Main function that accepts the configuration object passed by the decorator.
    It is based on the dummy Experiment subclass example.
    '''
    # Get the experiment parameters from the config object.
    params = config.experiment.params

    # Return some numerical value based on the parameters. In a real
    # experiment, this would be a validation or test score after training
    # and fitting a model.
    return params.x * params.y * params.z


if __name__ == '__main__':
    sys.exit(main())  # pylint: disable=no-value-for-parameter
